import sys

from pyqtgraph import functions as fn
from qtpy import QtCore, QtGui, QtWidgets
from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject

import numpy as np
from pymodaq_data import DataRaw, Axis, DataToExport

from pymodaq_gui.plotting.data_viewers.viewer2D import Viewer2D


class Curve(GraphicsObject):
    """
    **Bases:** :class:`GraphicsObject <pyqtgraph.GraphicsObject>`

    Item displaying an isocurve of a 2D array. To align this item correctly with an
    ImageItem, call ``isocurve.setParentItem(image)``.
    """

    def __init__(self, pen='w'):
        """
        Create a new isocurve item.

        ==============  ===============================================================
        **Arguments:**
        data            A 2-dimensional ndarray. Can be initialized as None, and set
                        later using :func:`setData <pyqtgraph.IsocurveItem.setData>`
        level           The cutoff value at which to draw the isocurve.
        pen             The color of the curve item. Can be anything valid for
                        :func:`mkPen <pyqtgraph.mkPen>`
        axisOrder       May be either 'row-major' or 'col-major'. By default this uses
                        the ``imageAxisOrder``
                        :ref:`global configuration option <apiref_config>`.
        ==============  ===============================================================
        """
        GraphicsObject.__init__(self)

        self.path: QtGui.QPainterPath = None
        self.pen: QtGui.QPen = None
        self.data = [(0, 3), (1, 2), (2, 4), (3.5, 0), (3.5, -1),]

        self.setPen(pen)

    def setData(self, data, level=None):
        """
        Set the data/image to draw isocurves for.

        ==============  ========================================================================
        **Arguments:**
        data            A 2-dimensional ndarray.
        level           The cutoff value at which to draw the curve. If level is not specified,
                        the previously set level is used.
        ==============  ========================================================================
        """

        self.data = data
        self.path = None
        self.prepareGeometryChange()
        self.update()

    def setPen(self, *args, **kwargs):
        """Set the pen used to draw the isocurve. Arguments can be any that are valid
        for :func:`mkPen <pyqtgraph.mkPen>`"""
        self.pen = fn.mkPen(*args, **kwargs)
        self.update()

    def setBrush(self, *args, **kwargs):
        """Set the brush used to draw the isocurve. Arguments can be any that are valid
        for :func:`mkBrush <pyqtgraph.mkBrush>`"""
        self.brush = fn.mkBrush(*args, **kwargs)
        self.update()

    def boundingRect(self):
        if self.data is None:
            return QtCore.QRectF()
        if self.path is None:
            self.generatePath()
        return self.path.boundingRect()

    def generatePath(self):
        self.path = QtGui.QPainterPath()
        self.path.moveTo(*self.data[0])
        for line in self.data[1:]:
            self.path.lineTo(*line)

    def paint(self, p, *args):
        if self.data is None:
            return
        if self.path is None:
            self.generatePath()
        p.setPen(self.pen)
        p.drawPath(self.path)


def main(data_distribution='uniform'):
    """either 'uniform' or 'spread'"""

    from pymodaq_gui.examples.curves import Curve

    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()

    widget_button = QtWidgets.QWidget()
    widget_button.setLayout(QtWidgets.QHBoxLayout())
    button = QtWidgets.QPushButton('New Data')
    ndata = QtWidgets.QSpinBox()
    widget_button.layout().addWidget(button)
    widget_button.layout().addWidget(ndata)

    def print_data(data: DataToExport):
        print(data)
        print('******')
        print(data.get_data_from_dim('Data1D'))

    data_to_plot = generate_uniform_data()


    prog = Viewer2D(widget)
    widget.show()
    prog.data_to_export_signal.connect(print_data)

    prog.view.get_action('histo').trigger()
    prog.view.get_action('autolevels').trigger()

    prog.show_data(data_to_plot)

    prog.view.show_roi_target(True)
    prog.view.move_scale_roi_target((50, 40), (10, 20))

    prog.show_data(data_to_plot)

    curve = Curve()
    prog.view.plotitem.addItem(curve)
    unscaled_x, unscaled_y = prog.view.unscale_axis([d[0] for d in curve.data], [d[1] for d in curve.data])
    curve.setData(list(zip(unscaled_x, unscaled_y)))


    QtWidgets.QApplication.processEvents()
    sys.exit(app.exec_())


def generate_uniform_data() -> DataRaw:
    from pymodaq_utils.math_utils import gauss2D
    Nx = 100
    Ny = 2 * Nx
    data_random = np.random.normal(size=(Ny, Nx))
    x = 0.5 * np.linspace(-Nx / 2, Nx / 2 - 1, Nx)
    y = 0.2 * np.linspace(-Ny / 2, Ny / 2 - 1, Ny)
    data_red = 3 * np.sin(x / 5) ** 2 * gauss2D(x, 5, Nx / 10, y, -1, Ny / 10, 1, 90) + 0.2 * data_random
    data_green = 10 * gauss2D(x, -20, Nx / 10, y, -10, Ny / 20, 1, 0)
    data_green[70:80, 7:12] = np.nan

    data_to_plot = DataRaw(name='mydata', distribution='uniform',
                                   data=[data_red, data_green, data_red-data_green],
                                   labels = ['myreddata', 'mygreendata'],
                                   axes=[Axis('xaxis', units='xpxl', data=x, index=1),
                                         Axis('yaxis', units='ypxl', data=y, index=0), ])
    return data_to_plot



if __name__ == '__main__':
    main()


from .dock import Dock, DockArea
from .list_picker import ListPicker
from .file_io import select_file
from .widgets.table import TableView, TableModel, SpinBoxDelegate, BooleanDelegate
from .widgets.push import PushButtonIcon, EditPush, EditPushRel, EditPushInfo
from .widgets.qled import QLED
from .widgets.spinbox import QSpinBox_ro, QSpinBoxWithShortcut
from .custom_app import CustomApp

from .utils import set_dark_palette


from qtpy import QtWidgets, QtGui, QtCore
from pyqtgraph.widgets.SpinBox import SpinBox


class SpinBox(SpinBox):
    """
    In case I want to add pyqtgraph spinbox functionalities
    """
    def __init__(self, *args, font_size=None, min_height=20, **kwargs):
        super().__init__(*args, **kwargs)

        if font_size is not None:
            self.set_font_size(font_size)
        self.setMinimumHeight(min_height)

    def set_font_size(self, font_size):
        font = QtGui.QFont()
        font.setPointSize(font_size)
        self.setFont(font)


class QSpinBoxWithShortcut(SpinBox):
    """
    QSpinBox but which accepts key sequences and store them as attribute
    For now, it does not apply to regular input such as text or numerics.

    Beware I could not make it run for the KeySequence Ctrl+Enter or any combination involving enter...

    """
    def __init__(self, *args, key_sequences=("Ctrl+E",), **kwargs):

        super().__init__(*args, **kwargs)

        self.shortcut = dict() #Store shortcuts in a dictionnary
        for key_sequence in key_sequences:
            shortcut = QtWidgets.QShortcut(QtGui.QKeySequence(key_sequence), self)
            shortcut.setContext(QtCore.Qt.WidgetWithChildrenShortcut)
            self.shortcut[key_sequence] = shortcut


class QSpinBox_ro(SpinBox):
    def __init__(self, *args, readonly=True, **kwargs):
        super().__init__(*args, **kwargs)
        #self.setMaximum(100000)
        self.setReadOnly(readonly)
        self.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)


if __name__ == '__main__':

    from pymodaq_gui.utils.utils import mkQApp

    app = mkQApp('Test Spinbox')

    spinbox = QSpinBoxWithShortcut(key_sequences=('Ctrl+E', 'Ctrl+Shift+E', 'Ctrl+Enter', 'Ctrl+Shift+Enter'))
    def print_spinbox(value):
        print(value)
    spinbox.shortcut['Ctrl+E'].activated.connect(lambda: print_spinbox(f'Ctrl+E: {spinbox.value()}'))
    spinbox.shortcut['Ctrl+Shift+E'].activated.connect(lambda: print_spinbox(f'Ctrl+Shift+E: {spinbox.value()}'))
    spinbox.shortcut['Ctrl+Shift+Enter'].activated.connect(lambda: print_spinbox(f'Ctrl+Shift+Enter: {spinbox.value()}'))
    spinbox.shortcut['Ctrl+Enter'].activatedAmbiguously.connect(lambda: print_spinbox(f'Ctrl+Enter: {spinbox.value()}'))
    spinbox.show()

    app.exec()


from .qled import QLED
from .push import PushButtonIcon, EditPush, EditPushRel
from .spinbox import QSpinBox_ro, SpinBox, QSpinBoxWithShortcut
from .table import TableView, TableModel
from .label import LabelWithFont



from pathlib import Path
from typing import Union

from qtpy import QtWidgets, QtGui


class WidgetWithBkg(QtWidgets.QWidget):
    """ Widget with a png file as a background texture

    Parameters
    ----------
    bkg_path: Path
        Path to a valid png file to be used as background
    """

    def __init__(self, bkg_path: Union[str, Path], *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(bkg_path, str):
            bkg_path = Path(bkg_path)
            if not bkg_path.is_file():
                raise ValueError(f'Unknown background file with path: {bkg_path}')

        self._bkg_path = bkg_path
        self.setup_palette()

    def setup_palette(self):
        pixmap = QtGui.QPixmap(str(self._bkg_path))
        self.setFixedSize(pixmap.size())

        palette = QtGui.QPalette()
        palette.setBrush(palette.ColorRole.Window, QtGui.QBrush(pixmap))
        self.setPalette(palette)
        self.setAutoFillBackground(True)



import os
import sys

from qtpy.QtCore import QObject, Signal, QEvent, QBuffer, QIODevice, Qt
from qtpy import QtWidgets, QtCore, QtGui

from pathlib import Path
from pymodaq_utils.config import Config
from pymodaq_utils.logger import set_logger, get_module_name
from pyqtgraph import mkQApp as mkQApppg


config = Config()
logger = set_logger(get_module_name(__file__))

def first_available_integer(liste):
    i = 0
    while i in liste:
        i += 1
    return i


here = Path(__file__).parent
custom_folder = here.parent.joinpath('QtDesigner_Ressources/custom/')
QtCore.QDir.addSearchPath('custom', str(custom_folder))


def set_dark_palette(app):
    from qtpy.QtGui import QPalette, QColor
    app.setStyle("Fusion")

    palette = QPalette()
    palette.setColor(QPalette.ColorRole.Window, QColor(53,53,53))
    palette.setColor(QPalette.ColorRole.WindowText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Base, QColor(42,42,42))
    palette.setColor(QPalette.ColorRole.ToolTipBase, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.ToolTipText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Dark, QColor(35,35,35))
    palette.setColor(QPalette.ColorRole.Shadow, QColor(20,20,20))
    palette.setColor(QPalette.ColorRole.AlternateBase, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ToolTipBase, Qt.GlobalColor.black)
    palette.setColor(QPalette.ColorRole.ToolTipText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Text, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.Button, QColor(53, 53, 53))
    palette.setColor(QPalette.ColorRole.ButtonText, Qt.GlobalColor.white)
    palette.setColor(QPalette.ColorRole.BrightText, Qt.GlobalColor.red)
    palette.setColor(QPalette.ColorRole.Link, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.Highlight, QColor(42, 130, 218))
    palette.setColor(QPalette.ColorRole.HighlightedText, QColor(250,250,250))
    palette.setColor(QPalette.Disabled, QPalette.ColorRole.ButtonText, Qt.GlobalColor.darkGray)
    palette.setColor(QPalette.Disabled, QPalette.ColorRole.WindowText, Qt.GlobalColor.darkGray)
    palette.setColor(QPalette.Disabled, QPalette.ColorRole.Text, Qt.GlobalColor.darkGray)
    palette.setColor(QPalette.Disabled, QPalette.ColorRole.Light, QColor(53, 53, 53))

    app.setPalette(palette)

    # Checkbox are not visible in dark style but it is not possible to
    # modify QCheckBox style without messing up the check mark (it disappears)
    # so images are needed to avoid the problem.
    app.setStyleSheet("""
        QCheckBox::indicator {
            width: 1em;
            height: 1em; 
        }
        QCheckBox::indicator:unchecked {
            image: url('custom:checkbox/unchecked.png');
        }
        QCheckBox::indicator:unchecked:disabled {
            image: url('custom:checkbox/unchecked_disabled.png');
        }
        QCheckBox::indicator:unchecked:focus {
            image: url('custom:checkbox/unchecked_focus.png');
        }
        QCheckBox::indicator:unchecked:pressed {
            image: url('custom:checkbox/unchecked_pressed.png');
        }
        QCheckBox::indicator:checked {
            image: url('custom:checkbox/checked.png');
        }
        QCheckBox::indicator:checked:disabled {
            image: url('custom:checkbox/checked_disabled.png');
        }
        QCheckBox::indicator:checked:focus {
            image: url('custom:checkbox/checked_focus.png');
        }
        QCheckBox::indicator:checked:pressed {
            image: url('custom:checkbox/checked_pressed.png');
        }
        QCheckBox::indicator:indeterminate {
            image: url('custom:checkbox/indeterminate.png');
        }
        QCheckBox::indicator:indeterminate:disabled {
            image: url('custom:checkbox/indeterminate_disabled.png');
        }
        QCheckBox::indicator:indeterminate:focus {
            image: url('custom:checkbox/indeterminate_focus.png');
        }
        QCheckBox::indicator:indeterminate:pressed {
            image: url('custom:checkbox/indeterminate_pressed.png');
        }

        QToolBarExtension {
            background: #555555;
            qproperty-icon: url('custom:arrow/right.png');
        }
        QToolTip {
            color: white;
            background-color: #555555;
            border: 1px solid white; 
        }
        """)

def clickable(widget):
    class Filter(QObject):
        clicked = Signal()

        def eventFilter(self, obj, event):
            if obj == widget:
                if event.type() == QEvent.MouseButtonRelease:
                    if obj.rect().contains(event.pos()):
                        self.clicked.emit()
                        # The developer can opt for .emit(obj) to get the object within the slot.
                        return True
            return False

    filter = Filter(widget)
    widget.installEventFilter(filter)
    return filter.clicked


def h5tree_to_QTree(base_node, base_tree_elt=None, pixmap_items=[]):
    """
        | Convert a loaded h5 file to a QTreeWidgetItem element structure containing two columns.
        | The first is the name of the h5 current node, the second is the path of the node in the h5 structure.
        |
        | Recursive function discreasing on base_node.

        ==================   ======================================== ===============================
        **Parameters**        **Type**                                 **Description**

          *h5file*            instance class File from tables module   loaded h5 file

          *base_node*         pytables h5 node                         parent node

          *base_tree_elt*     QTreeWidgetItem                          parent QTreeWidgetItem element
        ==================   ======================================== ===============================

        Returns
        -------
        QTreeWidgetItem
            h5 structure copy converted into QtreeWidgetItem structure.

        See Also
        --------
        h5tree_to_QTree

    """

    if base_tree_elt is None:
        base_tree_elt = QtWidgets.QTreeWidgetItem([base_node.name, "", base_node.path])
    for node_name, node in base_node.children().items():
        child = QtWidgets.QTreeWidgetItem([node_name, "", node.path])
        klass = node.attrs['CLASS']
        tooltip = []

        if 'origin' in node.attrs.attrs_name:
            tooltip.append(node.attrs['origin'])
        elif klass == 'GROUP':
            for c in node.children().values():
                if 'origin' in c.attrs.attrs_name:
                    tooltip.append(c.attrs['origin'])
                    break

        if hasattr(node, 'title') and node.title:
            tooltip.append(node.title)

        child.setToolTip(0, '/'.join(tooltip))

        if 'pixmap' in node.attrs.attrs_name:
            pixmap_items.append(dict(node=node, item=child))

        if klass == 'GROUP':
            h5tree_to_QTree(node, child, pixmap_items)

        base_tree_elt.addChild(child)
    return base_tree_elt, pixmap_items


def set_enable_recursive(children, enable=False):
    """Apply the enable state on all children widgets, do it recursively

    Parameters
    ----------
    children: (list) elements children ofa pyqt5 element
    enable: (bool) set enabled state (True) of all children widgets
    """
    for child in children:
        if not children:
            return
        elif isinstance(child, QtWidgets.QSpinBox) or isinstance(child, QtWidgets.QComboBox) or \
                isinstance(child, QtWidgets.QPushButton) or isinstance(child, QtWidgets.QListWidget):
            child.setEnabled(enable)
        else:
            set_enable_recursive(child.children(), enable)


def widget_to_png_to_bytes(widget, keep_aspect=True, width=200, height=100):
    """
    Renders the widget content in a png format as a bytes string
    Parameters
    ----------
    widget: (QWidget) the widget to render
    keep_aspect: (bool) if True use width and the widget aspect ratio to calculate the height
                        if False use set values of width and height to produce the png
    width: (int) the rendered width of the png
    height: (int) the rendered width of the png

    Returns
    -------
    binary string

    """
    png = widget.grab().toImage()
    wwidth = widget.width()
    wheight = widget.height()
    if keep_aspect:
        height = width * wheight / wwidth

    png = png.scaled(int(width), int(height), QtCore.Qt.KeepAspectRatio)
    buffer = QtCore.QBuffer()
    buffer.open(QtCore.QIODevice.WriteOnly)
    png.save(buffer, "png")
    return buffer.data().data()


def pngbinary2Qlabel(databinary, scale_height: int = None):
    buff = QBuffer()
    buff.open(QIODevice.WriteOnly)
    buff.write(databinary)
    dat = buff.data()
    pixmap = QtGui.QPixmap()
    pixmap.loadFromData(dat, 'PNG')
    if scale_height is not None and isinstance(scale_height, int):
        pixmap = pixmap.scaledToHeight(scale_height)
    label = QtWidgets.QLabel()
    label.setPixmap(pixmap)
    return label


def start_qapplication() -> QtWidgets.QApplication:
    app = QtWidgets.QApplication(sys.argv)
    if config('style', 'darkstyle'):
        set_dark_palette(app)
    return app


def mkQApp(name: str):
    app = mkQApppg(name)
    if config('style', 'darkstyle'):
        set_dark_palette(app)
    return app


def exec():
    app = mkQApp('a name')
    return app.exec() if hasattr(app, 'exec') else app.exec_()




import os
import sys
import warnings

from qtpy import QtCore
from qtpy.QtCore import QLocale
from pymodaq_utils import logger as logger_module
from pymodaq_utils.config import Config

from pymodaq_gui.qvariant import QVariant


logger = logger_module.set_logger(logger_module.get_module_name(__file__))

config = Config()


def decode_data(encoded_data):
    """
    Decode QbyteArrayData generated when drop items in table/tree/list view
    Parameters
    ----------
    encoded_data: QByteArray
                    Encoded data of the mime data to be dropped
    Returns
    -------
    data: list
            list of dict whose key is the QtRole in the Model, and the value a QVariant

    """
    data = []

    ds = QtCore.QDataStream(encoded_data, QtCore.QIODevice.ReadOnly)
    while not ds.atEnd():
        row = ds.readInt32()
        col = ds.readInt32()

        map_items = ds.readInt32()
        item = {}
        for ind in range(map_items):
            key = ds.readInt32()
            #TODO check this is fine
            value = QVariant()
            #value = None
            ds >> value
            item[QtCore.Qt.ItemDataRole(key)] = value.value()
        data.append(item)
    return data


def setLocale():
    """
    defines the Locale to use to convert numbers to strings representation using language/country conventions
    Default is English and US
    """
    language = getattr(QLocale, config('style', 'language'))
    country = getattr(QLocale, config('style', 'country'))
    QLocale.setDefault(QLocale(language, country))




from abc import abstractmethod
from typing import TYPE_CHECKING, List, Tuple, Union, Callable

import numpy as np
import pyqtgraph as pg
from pymodaq_data.post_treatment.process_to_scalar import DataProcessorFactory
from pymodaq_utils.logger import get_module_name, set_logger
from pymodaq_utils.enums import StrEnum
from pymodaq_utils.math_utils import rotate2D
from pymodaq_utils.utils import plot_colors
from pyqtgraph import ROI as pgROI
from pyqtgraph import LinearRegionItem as pgLinearROI
from pyqtgraph import functions as fn
from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import QSignalBlocker, Signal, Slot

from pymodaq_gui.config_saver_loader import get_set_roi_path
from pymodaq_gui.parameter import (Parameter, ParameterTree,
                                   )
from pymodaq_gui.plotting.utils import plot_utils
data_processors = DataProcessorFactory()

roi_path = get_set_roi_path()
logger = set_logger(get_module_name(__file__))
translate = QtCore.QCoreApplication.translate


ROI_NAME_PREFIX = 'ROI_'
def roi_format(index):
    return f'{ROI_NAME_PREFIX}{index:02d}'


class DataDim(StrEnum):
    Data1D = 'Data1D'
    Data2D = 'Data2D'


class ROIBase:
    """ Base class to be inherited for ROI to be created by the factory"""
    DIMENSIONALITY: DataDim = NotImplemented
    DESCRIPTOR: str = NotImplemented  # the identifier of the ROI, its name!


class ROIFactory():
    """The factory class for creating ROI"""

    registry = {}

    @classmethod
    def register(cls) -> Callable:
        """Class decorator method to register ROI class to the internal registry. Must be used as
        decorator above the definition of a ROI class.
        """

        def inner_wrapper(wrapped_class: ROIBase) -> ROIBase:
            if wrapped_class.DIMENSIONALITY   is NotImplemented or \
                    wrapped_class.DESCRIPTOR is NotImplemented:
                raise NotImplementedError(f'{wrapped_class} does not properly provide a valid value for '
                                          f'`DIMENSIONALITY` ({wrapped_class.DIMENSIONALITY}) or for '
                                          f'`ROI_DESC` ({wrapped_class.DESCRIPTOR})')

            if wrapped_class.DIMENSIONALITY not in cls.registry:
                cls.registry[wrapped_class.DIMENSIONALITY] = {}
            if wrapped_class.DESCRIPTOR not in cls.registry[wrapped_class.DIMENSIONALITY]:
                cls.registry[wrapped_class.DIMENSIONALITY][wrapped_class.DESCRIPTOR] = wrapped_class
            return wrapped_class
        return inner_wrapper

    @classmethod
    def create(cls, dimensionality: DataDim, descriptor: str, *args, **kwargs) -> ROIBase:
        """Factory command to create the ROI object.
        This method gets the appropriate ROI class from the registry and instantiates it.
        Parameters
        ----------
        dimensionality: DataDim
            the dimensionality of the ROI
        descriptor: str
            the roi descriptor string
        Returns
        -------
        an instance of the ROI created
        """
        if dimensionality not in cls.registry:
            raise ValueError(f".{dimensionality} is not a supported ROI dimensionality")
        elif descriptor not in cls.registry[dimensionality]:
            raise ValueError(f".{descriptor} is not a supported file description.")

        return cls.registry[dimensionality][descriptor](*args, **kwargs)

    @classmethod
    def get_dimensionality(cls):
        """Returns a list of registered dimensionality"""
        return list(cls.registry.keys()).sort()

    @classmethod
    def get_descriptors_from_dimensionality(cls, dim: DataDim):
        """Returns a list of ROi descriptors for a given dimensionality"""
        descriptors = list(cls.registry[dim].keys())
        descriptors.sort()
        return descriptors


class ROIMixin(QtCore.QObject):
    index_signal = Signal(int)

    def __init__(self, index=0, name='roi', compute=True):
        super().__init__()
        self.name = name
        self.index = index
        self._compute = compute
        self.menu = None

        self.signalBlocker = QSignalBlocker(self)
        self.signalBlocker.unblock()
        self._clipboard = QtGui.QGuiApplication.clipboard()

    def emit_index_signal(self):
        self.index_signal.emit(self.index)

    @abstractmethod
    def mouseClickEvent(self, ev):
        ...

    @abstractmethod
    def color(self):
        ...

    @abstractmethod
    def getMenu(self):
        ...

    def _emitCopyRequest(self):
        self.sigCopyRequested.emit(self)

    def mouseDoubleClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            self.sigDoubleClicked.emit(self, ev)

    @abstractmethod
    def copy_clipboard(self):
        ...

    @abstractmethod
    def center(self):
        ...

    @abstractmethod
    def width(self):
        ...
    @abstractmethod
    def height(self):
        ...

    def key(self) -> str:
        return roi_format(self.index)

    def type(self) -> str:
        return type(self).__name__

    def doShow(self, status: bool = True):
        if status:
            self.show()
        else:
            self.hide()

    @property
    def compute(self):
        return self._compute

    @compute.setter
    def compute(self, compute: bool = True):
        self._compute = compute


class ROI(pgROI, ROIMixin, ROIBase):
    """ Base class for all 2D ROI"""
    sigCopyRequested = Signal(object)
    sigDoubleClicked = Signal(object, object)
    sigRemoveRequested = Signal(object)

    def __init__(self, *args, index=0, name='roi', compute=True, **kwargs):
        ROIMixin.__init__(self, index=index, name=name, compute=compute)
        pgROI.__init__(self, *args, **kwargs)
        ROIBase.__init__(self)

    def getMenu(self):
        if self.menu is None:
            self.menu = QtWidgets.QMenu()
            self.menu.setTitle(translate("ROI", "ROI"))
            self.menu.addAction('Copy ROI to clipboard', self.copy_clipboard)
            self.menu.addAction("Copy ROI", self._emitCopyRequest)
            self.menu.addAction("Remove ROI", self._emitRemoveRequest)
        return self.menu

    def contextMenuEnabled(self):
        return True

    def raiseContextMenu(self, ev):
        menu = self.getMenu()
        menu = self.scene().addParentContextMenus(self, menu, ev)
        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(int(pos.x()), int(pos.y())))

    def contextMenuEvent(self, event):
        if self.menu is not None:
            self.menu.exec(event.screenPos())

    def mouseClickEvent(self, ev):
        super().mouseClickEvent(ev)
        if ev.button() == QtCore.Qt.MouseButton.RightButton and self.contextMenuEnabled():
            self.raiseContextMenu(ev)
            ev.accept()
        elif self.acceptedMouseButtons() & ev.button():
            ev.accept()
            self.sigClicked.emit(self, ev)
        elif ev.button() == QtCore.Qt.MouseButton.MiddleButton:
            ev.accept()
            self._emitRemoveRequest()
        else:
            ev.ignore()

    @property
    def color(self):
        return self.pen.color()

    def center(self) -> pg.Point:
        """ Get the center position of the ROI """
        return pg.Point(self.pos() + rotate2D(point =(self.width()/2,self.height()/2), angle=np.deg2rad(self.angle())))

    def set_center(self, center: Union[pg.Point, Tuple[float, float]]):
        """ Set the center position of the ROI """
        self.setPos(center - rotate2D(point =(self.width()/2,self.height()/2), angle=np.deg2rad(self.angle())))

    def copy_clipboard(self):
        info = plot_utils.RoiInfo.info_from_rect_roi(self)
        self._clipboard.setText(str(info.to_slices()))

    def width(self) -> float:
        return self.size().x()

    def height(self) -> float:
        return self.size().y()


class ROIBrushable(ROI):
    def __init__(self, brush=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if brush is None:
            brush = QtGui.QBrush(QtGui.QColor(0, 0, 255, 50))
        self.setBrush(brush)

    def setBrush(self, *br, **kargs):
        """Set the brush that fills the region. Can have any arguments that are valid
        for :func:`mkBrush <pyqtgraph.mkBrush>`.
        """
        self.brush = fn.mkBrush(*br, **kargs)
        self.currentBrush = self.brush

    def paint(self, p, opt, widget):
        # p.save()
        # Note: don't use self.boundingRect here, because subclasses may need to redefine it.
        r = QtCore.QRectF(0, 0, self.state['size'][0], self.state['size'][1]).normalized()

        p.setRenderHint(QtGui.QPainter.Antialiasing)
        p.setPen(self.currentPen)
        p.setBrush(self.currentBrush)
        p.translate(r.left(), r.top())
        p.scale(r.width(), r.height())
        p.drawRect(0, 0, 1, 1)
        # p.restore()


@ROIFactory.register()
class LinearROI(pgLinearROI, ROIMixin, ROIBase):
    sigCopyRequested = Signal(object)
    sigDoubleClicked = Signal(object,object)
    sigRemoveRequested = Signal(object)

    DIMENSIONALITY = DataDim.Data1D
    DESCRIPTOR = 'LinearROI'

    def __init__(self, index=0, pos=[0, 10], name = 'roi', compute=True, **kwargs):
        ROIMixin.__init__(self, index=index, name=name, compute=compute)
        pgLinearROI.__init__(self, values=pos, **kwargs)
        ROIBase.__init__(self)

    def getMenu(self):
        if self.menu is None:
            self.menu = QtWidgets.QMenu()
            self.menu.setTitle(translate("ROI", "ROI"))
            self.menu.addAction('Copy ROI to clipboard', self.copy_clipboard)
            self.menu.addAction("Copy ROI", self._emitCopyRequest)
            self.menu.addAction("Remove ROI", self._emitRemoveRequest)
        return self.menu

    def contextMenuEnabled(self):
        return True

    def raiseContextMenu(self, ev):
        menu = self.getMenu()
        menu = self.scene().addParentContextMenus(self, menu, ev)
        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(int(pos.x()), int(pos.y())))

    def contextMenuEvent(self, event):
        if self.menu is not None:
            self.menu.exec(event.screenPos())

    def mouseClickEvent(self, ev):
        super().mouseClickEvent(ev)
        if ev.button() == QtCore.Qt.MouseButton.RightButton and self.contextMenuEnabled():
            self.raiseContextMenu(ev)
            ev.accept()
        elif self.acceptedMouseButtons() & ev.button():
            ev.accept()
            self.sigClicked.emit(self, ev)
        elif ev.button() == QtCore.Qt.MouseButton.MiddleButton:
            ev.accept()
            self._emitRemoveRequest()
        else:
            ev.ignore()

    def copy_clipboard(self):
        info = plot_utils.RoiInfo.info_from_linear_roi(self)
        self._clipboard.setText(str(info.to_slices()))

    def pos(self) -> Tuple[float, float]:
        return self.getRegion()

    def center(self) -> float:
        pos = self.pos()
        return (pos[0] + pos[1]) / 2

    def setPos(self, pos: Tuple[int, int]):
        self.setRegion(pos)

    def setPen(self, color):
        self.setBrush(color)

    @property
    def color(self):
        return self.brush.color()


@ROIFactory.register()
class EllipseROI(ROI):
    """
    Elliptical ROI subclass with one scale handle and one rotation handle.


    ============== =============================================================
    **Arguments**
    pos            (length-2 sequence) The position of the ROI's origin.
    size           (length-2 sequence) The size of the ROI's bounding rectangle.
    **args         All extra keyword arguments are passed to ROI()
    ============== =============================================================

    """

    DIMENSIONALITY = DataDim.Data2D
    DESCRIPTOR = 'EllipseROI'

    def __init__(self, index=0, pos=[0, 0], size=[10, 10], **kwargs):
        # QtGui.QGraphicsRectItem.__init__(self, 0, 0, size[0], size[1])
        super().__init__(pos=pos, size=size, index=index, **kwargs)
        self.addRotateHandle([1.0, 0.5], [0.5, 0.5])
        self.addScaleHandle([0.5 * 2. ** -0.5 + 0.5, 0.5 * 2. ** -0.5 + 0.5], [0.5, 0.5])

    def getArrayRegion(self, arr, img=None, axes=(0, 1), **kwds):
        """
        Return the result of ROI.getArrayRegion() masked by the elliptical shape
        of the ROI. Regions outside the ellipse are set to 0.
        """
        # Note: we could use the same method as used by PolyLineROI, but this
        # implementation produces a nicer mask.
        if kwds["returnMappedCoords"]:
            arr, coords = pgROI.getArrayRegion(self, arr, img, axes, **kwds)
        else:
            arr = pgROI.getArrayRegion(self, arr, img, axes, **kwds)
        if arr is None or arr.shape[axes[0]] == 0 or arr.shape[axes[1]] == 0:
            return arr
        w = arr.shape[axes[0]]
        h = arr.shape[axes[1]]
        # generate an ellipsoidal mask
        mask = np.fromfunction(
            lambda x, y: (((x + 0.5) / (w / 2.) - 1) ** 2 + ((y + 0.5) / (h / 2.) - 1) ** 2) ** 0.5 < 1, (w, h))

        # reshape to match array axes
        if axes[0] > axes[1]:
            mask = mask.T
        shape = [(n if i in axes else 1) for i, n in enumerate(arr.shape)]
        mask = mask.reshape(shape)
        if kwds["returnMappedCoords"]:
            return arr * mask, coords
        else:
            return arr * mask

    def paint(self, p, opt, widget):
        r = self.boundingRect()
        p.setRenderHint(QtGui.QPainter.Antialiasing)
        p.setPen(self.currentPen)

        p.scale(r.width(), r.height())  # workaround for GL bug
        r = QtCore.QRectF(r.x() / r.width(), r.y() / r.height(), 1, 1)

        p.drawEllipse(r)

    def shape(self):
        self.path = QtGui.QPainterPath()
        self.path.addEllipse(self.boundingRect())
        return self.path


@ROIFactory.register()
class CircularROI(EllipseROI):

    DIMENSIONALITY = DataDim.Data2D
    DESCRIPTOR = 'CircularROI'

    def __init__(self, index=0, pos=[0, 0], size=[10, 10], **kwargs):
        ROI.__init__(self, pos=pos, size=size, index=index, **kwargs)
        self.addScaleHandle([0.5 * 2. ** -0.5 + 0.5, 0.5 * 2. ** -0.5 + 0.5], [0.5, 0.5],
                            lockAspect=True)


class SimpleRectROI(ROI):
    r"""
    Rectangular ROI subclass with a single scale handle at the top-right corner.
    """

    def __init__(self, pos=[0, 0], size=[10, 10], centered=False, sideScalers=False, **args):
        super().__init__(pos, size, **args)
        if centered:
            center = [0.5, 0.5]
        else:
            center = [0, 0]

        self.addScaleHandle([1, 1], center)
        if sideScalers:
            self.addScaleHandle([1, 0.5], [center[0], 0.5])
            self.addScaleHandle([0.5, 1], [0.5, center[1]])


@ROIFactory.register()
class RectROI(ROI):

    DIMENSIONALITY = DataDim.Data2D
    DESCRIPTOR = 'RectROI'

    def __init__(self, index=0, pos=[0, 0], size=[10, 10], **kwargs):
        super().__init__(pos=pos, size=size, index=index, **kwargs)  # , scaleSnap=True, translateSnap=True)
        self.addScaleHandle([1, 1], [0, 0])
        self.addRotateHandle([0, 0], [0.5, 0.5])


class ROIPositionMapper(QtWidgets.QWidget):
    """ Widget presenting a Tree structure representing a ROI positions.
    """

    def __init__(self, roi_pos, roi_size):
        super().__init__()
        self.roi_pos = roi_pos
        self.roi_size = roi_size

    def show_dialog(self):
        self.params = [
            {'name': 'position', 'type': 'group', 'children': [
                {'name': 'x0', 'type': 'float', 'value': self.roi_pos[0] + self.roi_size[0] / 2,
                 'step': 1},
                {'name': 'y0', 'type': 'float', 'value': self.roi_pos[1] + self.roi_size[1] / 2,
                 'step': 1}
            ]},
            {'name': 'size', 'type': 'group', 'children': [
                {'name': 'width', 'type': 'float', 'value': self.roi_size[0], 'step': 1},
                {'name': 'height', 'type': 'float', 'value': self.roi_size[1], 'step': 1}]
             }]

        dialog = QtWidgets.QDialog(self)
        vlayout = QtWidgets.QVBoxLayout()
        self.settings_tree = ParameterTree()
        vlayout.addWidget(self.settings_tree, 10)
        self.settings_tree.setMinimumWidth(300)
        self.settings = Parameter.create(name='settings', type='group', children=self.params)
        self.settings_tree.setParameters(self.settings, showTop=False)
        dialog.setLayout(vlayout)

        buttonBox = QtWidgets.QDialogButtonBox(parent=self)
        buttonBox.addButton('Apply', buttonBox.AcceptRole)
        buttonBox.accepted.connect(dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(dialog.reject)

        vlayout.addWidget(buttonBox)
        self.setWindowTitle('Set Precise positions for the ROI')
        res = dialog.exec()

        if res == dialog.Accepted:

            return self.settings
        else:
            return None




import importlib.util
import os
import sys
import pkgutil

def set_and_check_qt_backend_or_die(config):
    wanted_backend = config('qtbackend', 'backends')[0]
    backend = wanted_backend
    #filter to get only qt backend modules
    available_backends = [mod.name.lower() for mod in pkgutil.iter_modules() \
                          if mod.name.lower() in [ backend.lower() for backend in config('qtbackend', 'backends')]]

    backend_found = wanted_backend.lower() in available_backends
    if not backend_found:
        #trying in the remaining backends and taking the first one
        logger.warning(f"{backend} is not available. Trying to find another backend.")
        other_backends = [backend for backend in available_backends if backend != wanted_backend]
        if len(other_backends) > 0:
            backend_found = True
            backend =  other_backends[0]
            config['qtbackend']['backend'] = backend

    if backend_found:
        # environment variable is set
        os.environ['QT_API'] = backend
        try:
            import qtpy
            logger.info(f"{qtpy.API_NAME} Qt backend loaded")
        except ImportError as e:
            print(f'Should have selected {backend} for qtpy but still failed:')
            print(e)
            sys.exit(-1)

    else:
        msg = "No Qt backend could be found in your system, please install either pyqt6 or pyside6." \
              "(pyqt6 is preferred).\n"
        logger.error(msg)
        print(msg)
        sys.exit(-1)



try:
    from pymodaq_utils.utils import get_version, PackageNotFoundError
    __version__ = get_version(__package__)
except PackageNotFoundError:
    __version__ = '0.0.0dev'


from pymodaq_utils.config import Config
config = Config()  # to check for config file existence, otherwise create one

from pymodaq_utils.logger import set_logger
logger = set_logger('pymodaq_gui', base_logger=False)

logger.info('Starting PyMoDAQ GUI modules')
logger.info(f"Trying to set Qt backend to: {config['qtbackend']['backends'][0]}")
set_and_check_qt_backend_or_die(config)


from pymodaq_gui.qt_utils import setLocale

from pymodaq_data.plotting.plotter.plotter import register_plotter, PlotterFactory

logger.info(f"Setting Locale to {config['style']['language']} / {config['style']['country']}")
setLocale()

logger.info(f"Registering PyMoDAQ qt plotters...")

register_plotter(parent_module_name='pymodaq_gui.plotting.plotter')

logger.info(f"Done")

# in a try statement for compilation on readthedocs server but if this fail, you cannot use the code
from pymodaq_gui.plotting import data_viewers  # imported here as to avoid circular imports later on






import os
import sys
from typing import List, TYPE_CHECKING, Tuple, Union

from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import QObject, Slot, Signal,QSignalBlocker
from qtpy.QtGui import QIcon, QPixmap
from collections import OrderedDict

import pyqtgraph.Point as Point

from pymodaq_gui.parameter import utils as putils
from pymodaq_gui.parameter import ParameterTree, Parameter, ioxml, pymodaq_ptypes
from pyqtgraph.parametertree.parameterTypes.basetypes import GroupParameter

from pymodaq_gui.managers.action_manager import QAction

from pymodaq_utils.utils import plot_colors
from pymodaq_utils.logger import get_module_name, set_logger
from pymodaq_utils.config import Config
from pymodaq_gui.config_saver_loader import get_set_roi_path
from pymodaq_gui.utils import select_file
from pymodaq_gui.plotting.items.roi import ROIFactory, ROI, LinearROI, RectROI, DataDim


import numpy as np
from pathlib import Path
from pymodaq_data.post_treatment.process_to_scalar import DataProcessorFactory
from pymodaq_gui.utils.utils import first_available_integer

data_processors = DataProcessorFactory()

roi_path = get_set_roi_path()
logger = set_logger(get_module_name(__file__))
config = Config()


ROI_NAME_PREFIX = 'ROI_'
ROI2D_TYPES = ROIFactory.get_descriptors_from_dimensionality(DataDim.Data2D)


def roi_format(index):
    return f'{ROI_NAME_PREFIX}{index:02d}'


class ROIScalableGroup(GroupParameter):
    def __init__(self, roi_type = DataDim.Data1D, **opts):
        opts['type'] = 'group'
        opts['addText'] = "Add"
        self.roi_type = roi_type
        if roi_type == DataDim.Data2D:
            opts['addList'] = ROI2D_TYPES
        # self.color_list = ROIManager.color_list
        super().__init__(**opts)

    def addNew(self, typ=''):
        name_prefix = ROI_NAME_PREFIX
        child_indexes = [int(par.name()[len(name_prefix) + 1:]) for par in self.children()]
        if not child_indexes:
            newindex = 0
        else:
            newindex = max(child_indexes) + 1

        self.addChild(self.makeChild(newindex, typ))


    def makeChild(self, index, descriptor: str):
        child = {'name': ROIManager.roi_format(index), 'type': 'bool','value':True, 'removable': True, 'renamable': False, 'expanded': False,'context':['Copy',]}
        if self.roi_type == DataDim.Data2D:
            child['children'] = ROIScalableGroup.make_ROIParam2D(descriptor, index)
        elif self.roi_type == DataDim.Data1D:
            child['children'] = ROIScalableGroup.make_ROIParam1D(descriptor, index)
        return child  
    
    @staticmethod
    def makeChannelsParam(dim=DataDim.Data2D):
        if dim == DataDim.Data2D:
            child = [{'title': 'Use channel', 'name': 'use_channel', 'type': 'itemselect', 'checkbox': True,
                      'value': dict(all_items=['red', 'green', 'blue'],
                           selected=['red',]),
                              },]
        else:
            child = [{'title': 'Use channel', 'name': 'use_channel', 'type': 'itemselect','checkbox': True},]
        return child 

    @staticmethod
    def makeDisplayParam(index):
        return [{'name': 'Color', 'type': 'color', 'value': list(np.roll(ROIManager.color_list, index)[0])},
             {'name': 'zlevel', 'title':'Z-level','type': 'int', 'expanded': False, 'value':10},] 
        
    @staticmethod
    def makeMathParam(dim=DataDim.Data2D):
        return [{'title': 'Math type:', 'name': 'math_function', 'type': 'list',
                             'limits': data_processors.functions_filtered(dim)},]
    @staticmethod    
    def make_ROIParam2D(descriptor: str, index):
            children = []    
            children.extend([{'title': 'Type', 'name': 'roi_type', 'type': 'list', 'value': descriptor,
                              'limits': ROI2D_TYPES, 'readonly': False,}])
            children.append({'title': 'Process data', 'name': 'process_data', 'type': 'led_push',
                             'value': config.get(('plotting', 'process_roi'), True),})
            children.extend(ROIScalableGroup.makeChannelsParam(DataDim.Data2D))
            children.extend(ROIScalableGroup.makeMathParam(DataDim.Data2D))
            children.extend(ROIScalableGroup.makeDisplayParam(index))
            children.extend([{'name': 'center', 'type': 'group', 'expanded': False, 'children': [
                    {'name': 'x', 'type': 'float', 'value': 0, 'step': 1,'decimals':6},
                    {'name': 'y', 'type': 'float', 'value': 0, 'step': 1,'decimals':6}
                ]}, ])                
            children.extend([{'name': 'position', 'type': 'group', 'expanded': False, 'children': [
                    {'name': 'x', 'type': 'float', 'value': 0, 'step': 1,'decimals':6},
                    {'name': 'y', 'type': 'float', 'value': 0, 'step': 1,'decimals':6}
                ]}, ])          
            children.extend([
                    {'name': 'size', 'type': 'group', 'expanded': False, 'children': [
                        {'name': 'width', 'type': 'float', 'value': 10, 'step': 1,'decimals':6},
                        {'name': 'height', 'type': 'float', 'value': 10, 'step': 1,'decimals':6}
                    ]},
                    {'name': 'angle', 'type': 'float', 'value': 0, 'step': 1}])    
            return children

    @staticmethod    
    def make_ROIParam1D(descriptor: str, index):
            children = []
            children.append({'title': 'Process data', 'name': 'process_data', 'type': 'led_push',
                             'value': config.get(('plotting', 'process_roi'), True),})
            children.extend(ROIScalableGroup.makeChannelsParam(DataDim.Data1D))
            children.extend(ROIScalableGroup.makeMathParam(DataDim.Data1D))
            children.extend(ROIScalableGroup.makeDisplayParam(index))
            children.extend([{'name': 'position', 'type': 'group', 'children': [
                {'name': 'left', 'type': 'float', 'value': 0, 'step': 1},
                {'name': 'right', 'type': 'float', 'value': 10, 'step': 1}
                    ]}, ])
            
            return children



class ROIManager(QObject):

    new_ROI_signal = Signal(str)
    remove_ROI_signal = Signal(str)
    roi_value_changed = Signal(str, tuple)
    color_signal = Signal(list)
    roi_update_children = Signal(list)
    roi_changed = Signal()
    color_list = np.array(plot_colors)

    def __init__(self, viewer_widget=None, ROI_type=DataDim.Data1D):
        super().__init__()
        self.ROI_type = ROI_type
        self.roiwidget = QtWidgets.QWidget()
        self.viewer_widget = viewer_widget  # either a PlotWidget or a ImageWidget
        self._ROIs: OrderedDict[str, ROI] = OrderedDict([])
        self.setupUI()

    @staticmethod
    def roi_format(index):
        logger.warning(f'ROIManager.roi_format is deprecated, use roi_format')
        return roi_format(index)
    
    @property
    def ROIs(self):
        return self._ROIs
    
    def __len__(self):
        return len(self._ROIs)

    def get_roi_from_index(self, index: int) -> ROI:
        return self.ROIs[roi_format(index)]

    def _set_roi_from_index(self, index: int, roi):
        self.ROIs[roi_format(index)] = roi

    def get_roi(self, roi_key):
        if roi_key in self.ROIs:
            return self.ROIs[roi_key]
        else:
            raise KeyError(f'{roi_key} is not a valid ROI identifier for {self.ROIs}')

    def emit_colors(self):
        self.color_signal.emit([self._ROIs[roi_key].color for roi_key in self._ROIs])

    def add_roi_programmatically(self, descriptor: str = ROI2D_TYPES[0]):
        self.settings.child('ROIs').addNew(descriptor)

    def remove_roi_programmatically(self, index: int):
        self.settings.child('ROIs').removeChild(self.settings.child('ROIs', roi_format(index)))

    def setupUI(self):

        vlayout = QtWidgets.QVBoxLayout()
        self.roiwidget.setLayout(vlayout)

        self.toolbar = QtWidgets.QToolBar()
        vlayout.addWidget(self.toolbar)

        self.save_ROI_pb = QAction(QIcon(QPixmap("icons:save_ROI.png")), 'Save ROIs')
        self.load_ROI_pb = QAction(QIcon(QPixmap("icons:load_ROI.png")), 'Load ROIs')
        self.clear_ROI_pb = QAction(QIcon(QPixmap("icons:clear_ROI.png")), 'Clear ROIs')
        self.toolbar.addActions([self.save_ROI_pb, self.load_ROI_pb, self.clear_ROI_pb])


        self.roitree = ParameterTree()
        vlayout.addWidget(self.roitree)
        self.roiwidget.setMinimumWidth(250)
        self.roiwidget.setMaximumWidth(300)

        params = [
            {'title': 'Measurements:', 'name': 'measurements', 'type': 'table', 'value': OrderedDict([]), 'Ncol': 2,
             'header': ["LO", "Value"]},
            ROIScalableGroup(roi_type=self.ROI_type, name="ROIs")]
        self.settings = Parameter.create(title='ROIs Settings', name='rois_settings', type='group', children=params)
        self.roitree.setParameters(self.settings, showTop=False)
        self.settings.sigTreeStateChanged.connect(self.roi_tree_changed)
        self.settings_signalBlocker = QSignalBlocker(self.settings)
        self.settings_signalBlocker.unblock()

        self.save_ROI_pb.triggered.connect(self.save_ROI)
        self.load_ROI_pb.triggered.connect(lambda: self.load_ROI(None))
        self.clear_ROI_pb.triggered.connect(self.clear_ROI)

    def get_ROI_indexes(self,):
        return [roi.index for roi in self.ROIs.values()]

    def roi_tree_changed(self, param, changes):

        for param, change, data in changes:
            path = self.settings.childPath(param)
            if path is not None:
                childName = '.'.join(path)
            else:
                childName = param.name()
            if change == 'childAdded':  # new roi to create
                par: Parameter = data[0]
                roi = self.make_ROI(par)                
                self.add_ROI(roi)
                self.emit_colors()
                self.roi_changed.emit()

            elif change == 'value':
                if param.name() in putils.iter_children(self.settings.child('ROIs'), []):
                    parent_name = putils.get_param_path(param)[putils.get_param_path(param).index('ROIs')+1]
                    if parent_name in self._ROIs.keys():
                        roi_changed = self._ROIs[parent_name]                                    
                        self.update_roi(roi_changed, param)
                    self.roi_value_changed.emit(parent_name, (param, param.value()))
                if param.name() == 'Color':
                    self.emit_colors()
            elif change == 'parent':
                if 'ROI' in param.name():
                    self.remove_ROI(self.ROIs[param.name()])

            elif change == 'contextMenu':  # MenuSel
                if data=='Copy':
                    self.copy_ROI(self.ROIs[param.name()])                    

    def make_ROI(self, param: Parameter):
        newindex = int(param.name()[-2:])
        pos = self.viewer_widget.plotItem.vb.viewRange()
        if self.ROI_type == DataDim.Data1D:
            descriptor = ''
            pos = pos[0]
            pos = pos[0] + np.diff(pos)*np.array([2,4])/6
            roi = self.make_ROI1D(newindex, pos, brush=param['Color'],
                                  compute=param['process_data'])
        elif self.ROI_type == DataDim.Data2D:
            descriptor = param.child('roi_type').value()
            xrange,yrange=pos                    
            width = np.max(((xrange[1] - xrange[0]) / 10, 2))
            height = np.max(((yrange[1] - yrange[0]) / 10, 2))
            pos = [int(np.mean(xrange) - width / 2), int(np.mean(yrange) - width / 2)]
            roi = self.make_ROI2D(descriptor, index=newindex, pos=pos,size=[width, height],
                                  pen=param['Color'], compute=param['process_data'])

        return roi

    def add_ROI(self, roi):
        # Connection roi signals to relevant function
        roi.sigRegionChangeFinished.connect(lambda: self.roi_changed.emit())
        roi.sigRegionChangeFinished.connect(self.update_roi_tree)
        roi.sigRemoveRequested.connect(self.remove_ROI)
        roi.sigCopyRequested.connect(self.copy_ROI)        
        roi.setAcceptedMouseButtons(QtCore.Qt.MouseButton.LeftButton) 
        roi.sigDoubleClicked.connect(self.expand_roi_tree)
        # Updating tree
        self.update_roi_tree(roi)
        # Adding to dictionnary
        self.ROIs[roi.key()] = roi
        # Adding to viewer
        self.viewer_widget.plotItem.addItem(roi)  
        # Emitting signal
        self.new_ROI_signal.emit(roi.key())

    def expand_roi_tree(self, roi,):
        # Expand roi tree when roi gets double selected
        param = self.settings.child(*('ROIs', roi_format(roi.index)))
        isExpanded = not param.opts['expanded']
        param.setOpts(expanded=isExpanded)

    def make_ROI1D(self, index, pos, compute=True, **kwargs):
        """Convenience function to make custom ROI_1D

        Args:
            index (int): Current index of ROI
            pos: Initial position of ROI

        Returns:
            roi: LinearROI
        """
        roi = ROIFactory.create(DataDim.Data1D,
                                ROIFactory.get_descriptors_from_dimensionality(DataDim.Data1D)[0],
                                index=index, pos=pos, compute=compute, **kwargs)
        # roi.setZValue(-10)
        roi.setOpacity(0.2)
        return roi                    

    def make_ROI2D(self, descriptor: str, index, pos, size, compute=True, **kwargs):
        """Convenience function to make custom ROI_2D

        Args:
            descriptor (str): name of 2D ROI
            index (int): Current index of ROI
            pos: Initial position of ROI
            size: Initial size of ROI

        Returns:
            roi: pg.ROI 
        """

        return ROIFactory.create(DataDim.Data2D, descriptor,
                                 index=index, pos=pos,
                                 size=size, name=roi_format(index),
                                 compute=compute, **kwargs)
    
    def remove_ROI(self, roi):
        """Function to remove roi from dict and widget

        Args:
            roi (pg.ROI): roi to be removed
        """
        roi_group = self.settings.child('ROIs')
        for param in roi_group.children():                
                if roi.key() == param.name():
                    self.settings_signalBlocker.reblock()
                    roi_group.removeChild(param)
                    self.settings_signalBlocker.unblock()
        roi = self.ROIs.pop(roi.key())
        self.viewer_widget.plotItem.removeItem(roi)
        self.remove_ROI_signal.emit(roi.key())
        self.emit_colors()

    def copy_ROI(self, roi: ROI):
        """Method to copy a ROI and add it to the parameter tree and to the viewer widget
        The method extracts the parameters of the copied ROI, create a new parameter, a new ROI and update it with the settings from the copied parameter
        Args:
            roi (ROI): the ROI to be copied
        """
        index = first_available_integer(self.get_ROI_indexes()) 
        
        roi_group = self.settings.child('ROIs')
        #Copy parameter and edit name
        param_roi = self.get_parameter(roi)
        param = param_roi.saveState() # Transforming parameter in dict
        param['name'] = roi_format(index) # Changing name   
        param = Parameter.create(**param) # Transforming dict in parameter
        self.settings_signalBlocker.reblock()
        roi_group.addChild(param)
        self.settings_signalBlocker.unblock()
        new_roi = self.make_ROI(param)

        param_to_update = putils.iter_children_params(param_roi,[],filter_name=('roi_type',),filter_type=('group',)) # Parameters to update
        # [self.update_roi(new_roi,p) for p in reversed(param_to_update)]     
        self.add_ROI(new_roi)
        [self.update_roi(new_roi,p) for p in reversed(param_to_update)]     


    def update_use_channel(self, channels: List[str], index=None):
        """Function to update the selected channels. If no index is given, the channels are applied to all ROIs.

        Args:
            channels (List[str]): channels list from a viewer
            index (int, optional): ROI index. Defaults to None.
        """
        if index is not None:   
            param = self.settings.child('ROIs', roi_format(index), 'use_channel')
            param.setValue(dict(all_items=channels,
                        selected=channels))
        else:
            for ind in range(len(self)): 
                param = self.settings.child('ROIs', roi_format(ind), 'use_channel')
                param.setValue(dict(all_items=channels,
                        selected=channels))   
                    
    def update_roi(self, roi: ROI, param: Parameter):
        par = self.get_parameter(roi)
        roi.signalBlocker.reblock()
        parent_name = param.parent().opts['name']

        if param.name() == roi.key():
            roi.doShow(param.value())
        elif param.name() == 'roi_type':
            state = roi.saveState()
            self.viewer_widget.plotItem.removeItem(roi)            
            if self.ROI_type =='2D':
                roi = self.make_ROI2D(roi_type=param.value(),index=roi.index,pos=state['pos'],size=state['size'],angle=state['angle'],pen=roi.pen)                
                self.add_ROI(roi)
        elif param.name() == 'Color':
            roi.setPen(param.value())
            self.emit_colors()
        elif parent_name == 'center':
            center = roi.center()
            pos = self.update_roi_pos(center, param)
            if self.ROI_type =='1D':
                roi.set_positions()
                pos.sort()
            else:
                roi.set_center(pos)
        elif parent_name == 'position':
            position = roi.pos()
            pos = self.update_roi_pos(position, param)
            if self.ROI_type =='1D':
                pos = np.sort(pos) #Subclass pg.Point to implement sort?
                roi.setPos(pos) 
                self.settings_signalBlocker.reblock()
                par.child(*('position', 'left')).setValue(pos[0])
                par.child(*('position', 'right')).setValue(pos[1])         
                self.settings_signalBlocker.unblock()
            roi.setPos(pos)          
        elif param.name() == 'angle':
            roi.setAngle(param.value(),center=[0.5,0.5])
        elif param.name() == 'zlevel':
            roi.setZValue(param.value())
        elif param.name() == 'width':
            size = roi.size()
            roi.setSize((param.value(), size[1]))
        elif param.name() == 'height':
            size = roi.size()
            roi.setSize((size[0], param.value()))
        elif param.name() == 'process_data':
            roi.compute = param.value()

        self.update_roi_tree(roi)
        roi.signalBlocker.unblock()

    def update_roi_pos(self, pos, param):
        if param.name() == 'x' or param.name() == 'left':
            poss = Point(param.value(), pos[1])
        elif param.name() == 'y' or param.name() == 'right':         
            poss = Point(pos[0], param.value())                   
        return poss
    
    def get_parameter(self, roi):
        if type(roi) is int:
            par =  self.settings.child(*('ROIs', roi_format(roi)))
        else:
            par = self.settings.child(*('ROIs', roi.key()))
        return par

    @Slot(type(ROI))
    def update_roi_tree(self, roi):
        par = self.get_parameter(roi)        

        if isinstance(roi, LinearROI):
            pos = roi.getRegion()
        else:
            pos = roi.pos()
            size = roi.size()
            angle = roi.angle()
            center = roi.center()
            Zvalue = roi.zValue()

        self.settings_signalBlocker.reblock()
        if isinstance(roi, LinearROI):
            par.child(*('position', 'left')).setValue(pos[0])
            par.child(*('position', 'right')).setValue(pos[1])
        if not isinstance(roi, LinearROI):
            par.child(*('position', 'x')).setValue(pos.x())
            par.child(*('position', 'y')).setValue(pos.y())
            par.child(*('center', 'x')).setValue(center.x())
            par.child(*('center', 'y')).setValue(center.y())        
            par.child(*('size', 'width')).setValue(size.x())
            par.child(*('size', 'height')).setValue(size.y())
            par.child('angle').setValue(angle)
            par.child('zlevel').setValue(Zvalue)
        self.settings_signalBlocker.unblock()

    def save_ROI(self):

        try:
            data = ioxml.parameter_to_xml_string(self.settings.child(('ROIs')))
            path = select_file(start_path=Path.home(), ext='xml', save=True, force_save_extension=True)

            if path != '':
                with open(path, 'wb') as f:
                    f.write(data)
        except Exception as e:
            print(e)

    def clear_ROI(self):
        keys = [roi.key() for roi in self._ROIs.values()]
        for roi_key in keys:
            self.settings.child(*('ROIs', roi_key)).remove()

    def load_ROI(self, path=None, params=None):
        try:
            if params is None:
                if path is None:
                    path = select_file(start_path=Path.home(), save=False, ext='xml', filter='XML files (*.xml)')
                    if path != '':
                        params = Parameter.create(title='Settings', name='settings', type='group',
                                                  children=ioxml.XML_file_to_parameter(path))

            if params is not None:
                self.clear_ROI()
                QtWidgets.QApplication.processEvents()

                for param in params:
                    if 'roi_type' in putils.iter_children(param, []):
                        self.settings.child('ROIs').addNew(param.child('roi_type').value())
                    else:
                        self.settings.child('ROIs').addNew()
                QtWidgets.QApplication.processEvents()
                self.set_roi(self.settings.child('ROIs').children(), params)
        except Exception as e:
            logger.exception(str(e))

    def set_roi(self, roi_params, roi_params_new):
        for child, new_child in zip(roi_params, roi_params_new):
            if new_child.value():
                child.setValue(new_child.value())
            self.set_roi(child.children(), new_child.children())


class ROISaver:
    def __init__(self, msgbox=False, det_modules=[]):

        self.roi_presets = None
        self.detector_modules = det_modules

        if msgbox:
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText("ROI Manager?")
            msgBox.setInformativeText("What do you want to do?")
            cancel_button = msgBox.addButton(QtWidgets.QMessageBox.StandardButton.Cancel)
            modify_button = msgBox.addButton('Modify', QtWidgets.QMessageBox.ButtonRole.AcceptRole)
            msgBox.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Cancel)
            ret = msgBox.exec()

            if msgBox.clickedButton() == modify_button:
                path = select_file(start_path=roi_path, save=False, ext='xml')
                if path != '':
                    self.set_file_roi(str(path))
            else:  # cancel
                pass

    def set_file_roi(self, filename, show=True):
        """

        """

        children = ioxml.XML_file_to_parameter(filename)
        self.roi_presets = Parameter.create(title='roi', name='rois', type='group', children=children)

        det_children = [child for child in self.roi_presets.children() if 'det' in child.opts['name']]
        det_names = [child.child('detname').value() for child in self.roi_presets.children() if
                     'det' in child.opts['name']]
        det_module_names = [det.title for det in self.detector_modules]
        for ind_det, det_roi in enumerate(det_children):
            det_module = self.detector_modules[det_module_names.index(det_names[ind_det])]
            viewer_children = [child for child in det_roi.children() if 'viewer' in child.opts['name']]
            for ind_viewer, viewer in enumerate(det_module.viewers):
                rois_params = [child for child in viewer_children[ind_viewer].children() if 'ROI' in child.opts['name']]
                if len(rois_params) > 0:
                    if hasattr(viewer, 'roi_manager'):
                        if hasattr(viewer, 'activate_roi'):  # because for viewer 0D it is irrelevant
                            viewer.activate_roi()
                        viewer.roi_manager.load_ROI(params=rois_params)
                        QtWidgets.QApplication.processEvents()

        if show:
            self.show_rois()

    def set_new_roi(self, file=None):
        if file is None:
            file = 'roi_default'

        self.roi_presets = Parameter.create(name='roi_settings', type='group', children=[
            {'title': 'Filename:', 'name': 'filename', 'type': 'str', 'value': file}, ])

        for ind_det, det in enumerate(self.detector_modules):
            det_param = Parameter.create(name=f'det_{ind_det:03d}', type='group', children=[
                {'title': 'Det Name:', 'name': 'detname', 'type': 'str', 'value': det.title}, ])

            for ind_viewer, viewer in enumerate(det.ui.viewers):
                viewer_param = Parameter.create(
                    name=f'viewer_{ind_viewer:03d}', type='group',
                    children=[
                        {'title': 'Viewer:', 'name': 'viewername', 'type': 'str',
                         'value': det.ui.viewer_docks[ind_viewer].name()}, ])

                if hasattr(viewer, 'roi_manager'):
                    viewer_param.addChild(
                        {'title': 'ROI type:', 'name': 'roi_type', 'type': 'str',
                         'value': viewer.roi_manager.settings.child('ROIs').roi_type})
                    viewer_param.addChildren(viewer.roi_manager.settings.child('ROIs').children())
                det_param.addChild(viewer_param)
            self.roi_presets.addChild(det_param)

        ioxml.parameter_to_xml_file(self.roi_presets, os.path.join(roi_path, file))
        self.show_rois()

    def show_rois(self):
        """

        """
        dialog = QtWidgets.QDialog()
        vlayout = QtWidgets.QVBoxLayout()
        tree = ParameterTree()
        tree.setMinimumWidth(400)
        tree.setMinimumHeight(500)
        tree.setParameters(self.roi_presets, showTop=False)

        vlayout.addWidget(tree)
        dialog.setLayout(vlayout)
        buttonBox = QtWidgets.QDialogButtonBox(parent=dialog)

        buttonBox.addButton('Save', buttonBox.AcceptRole)
        buttonBox.accepted.connect(dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(dialog.reject)

        vlayout.addWidget(buttonBox)
        dialog.setWindowTitle('Fill in information about this manager')
        res = dialog.exec()

        if res == QtWidgets.QDialog.DialogCode.Accepted:
            # save managers parameters in a xml file
            # start = os.path.split(os.path.split(os.path.realpath(__file__))[0])[0]
            # start = os.path.join("..",'daq_scan')
            ioxml.parameter_to_xml_file(
                self.roi_presets, os.path.join(
                    roi_path, self.roi_presets.child('filename').value()))


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    from pymodaq_gui.plotting.widgets import ImageWidget
    from pyqtgraph import PlotWidget

    im = ImageWidget()
    im = PlotWidget()
    prog = ROIManager(im, DataDim.Data2D)
    widget = QtWidgets.QWidget()
    layout = QtWidgets.QHBoxLayout()
    widget.setLayout(layout)
    layout.addWidget(im)
    layout.addWidget(prog.roiwidget)
    widget.show()
    prog.add_roi_programmatically(ROI2D_TYPES[0])
    prog.add_roi_programmatically(ROI2D_TYPES[1])
    prog.add_roi_programmatically(ROI2D_TYPES[2])
    sys.exit(app.exec_())


import argparse
from pathlib import Path
import sys
import os
from qtpy import QtWidgets

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
from pymodaq_gui.h5modules.browsing import H5Browser
from pymodaq_utils.config import Config


config = Config()


def main(h5file_path: Path = None):
    from pymodaq_gui.utils.utils import mkQApp
    import sys
    app = mkQApp('H5Browser')

    h5file_path_tmp = None
    parser = argparse.ArgumentParser(description="Opens HDF5 files and navigate their contents")
    parser.add_argument("-i", "--input", help="specify path to the file to be opened")
    args = parser.parse_args()

    if args.input:
        h5file_path_tmp = Path(args.input).resolve()  # Transform to absolute Path in case it is relative

        if not h5file_path_tmp.exists():
            print(f'Error: {args.input} does not exist. Opening h5browser without input file.')
            h5file_path_tmp = h5file_path

    win = QtWidgets.QMainWindow()
    prog = H5Browser(win, h5file_path=h5file_path_tmp)
    win.show()
    QtWidgets.QApplication.processEvents()

    app.exec()


if __name__ == '__main__':
    main()
